% Provides a script to reproduce the benchmark of the paper. 
% 
% Over 10 random splits (90% cross-validation, 10% testing), it performs 10-fold cross-validation driven by the area under precision-recall curve.
%
% Copyright (C) 2012 Rodolphe Jenatton, Nicolas Le Roux, Antoine Bordes, Guillaume Obozinski
%         
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details. 
% 
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.



clear all
rand('state',0);
randn('state',0);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% dataset choice
dataset = 'dnations_extended';

switch dataset
case 'alyawarra'
	load('alyawarradata.mat');
case {'dnations','dnations_extended'}
	load([dataset '.mat']);
	Rs = R;
case 'uml'
	load('uml.mat');
end

triplets = zeros(numel(Rs), 3);
related  = zeros(numel(Rs), 1);

% Transform the 3D-tensor into a matrix.
index = 0;
for i = 1:size(Rs, 1)
	for j = 1:size(Rs, 2)
		for k = 1:size(Rs, 3)
			% There might be NaNs.
			if ~isnan(Rs(i,j,k))
				index = index + 1;
				triplets(index, :) = [i k j];
				related(index, 1) = Rs(i,j,k);
			end
		end
	end
end

% Remove unused rows (which might be there because of the NaN).
triplets = triplets(1:index, :);
related  = related(1:index, :);

n_relationships = size(Rs, 3);
n_subjects      = size(Rs, 1);
n_objects       = n_subjects;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Setting of the parameters which lie outside of the cross validation procedure
ratio_train = 0.9;
ratio_test  = 1 - ratio_train;
n_splits    = 10;
k_folds     = 10;

params.model.ngram = '111';

% Based on computational cost considerations, the number of maximum iterations can be varied in the training/validation and testing pahses.
max_iter_train = 50;
max_iter_test  = 130;

params.opt.batchsize  = 100;
params.opt.max_iter   = max_iter_train;
params.opt.rate       = .1;
params.opt.decr       = 0;
params.opt.proj_skip  = 1;

params.misc.save = 0;

params.model.unique_rep        = 1;
params.model.constraint_for_SO = 0;

params.neg_sampling.sample_type = 'rel';
params.neg_sampling.n_neg       = 0;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
neg_weight_range = [0.5];% Results may probably be improved by cross-validating over this parameter

alpha_range      = [0.05, 0.1, 0.5, 1];
n_factors_range  = [100, 200, 500];
n_features_range = [10, 25, 50];

%
% Best parameters for each dataset
%
%switch dataset
%case 'alyawarra'
%	alpha_range      = [0.5];
%	n_factors_range  = [500];
%	n_features_range = [50];
%case {'dnations','dnations_extended'}
%	alpha_range      = [0.1];
%	n_factors_range  = [100];
%	n_features_range = [10];
%case 'uml'
%	alpha_range      = [1];
%	n_factors_range  = [200];
%	n_features_range = [10];
%end

CV_scores       = zeros(k_folds, length(alpha_range), length(n_factors_range), length(n_features_range), length(neg_weight_range));
rd_split_scores = zeros(n_splits, 2);
best_params     = zeros(n_splits, 4);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for s=1:n_splits,

	% Create a random split 
	rd_idx         = randperm(rows(triplets));
	train_triplets = triplets( rd_idx, :);
	train_related  = related( rd_idx );

	n_train = floor(ratio_train * rows(triplets));

	test_triplets  = train_triplets((n_train+1):end, :);
	train_triplets = train_triplets(1:n_train, :);	

	test_related  = train_related((n_train+1):end);
	train_related = train_related(1:n_train);	

	% We need to make sure that the each cross-validation split contains
	% each subject, relationship and object at least once.
	tries = 0;
	valid = 0;

	while valid == 0,

		[train_idx, test_idx] = get_cross_valid_idx(n_train, k_folds);

		valid = check_CV_splits_validity( train_idx, train_triplets, n_relationships, n_subjects );

		tries = tries + 1;

		if tries > 200, fprintf('Too many tries, aborted.\n'); return; end

	end

	% We learn a model over each fold, for each combination of parameters
	for ifold=1:k_folds,

		fprintf('\t*** %d-th random split (out of %d): %d-th fold (out of %d) being processed ***\n',s,n_splits,ifold,k_folds);

		for ialpha=1:length(alpha_range),
			for ifactor=1:length(n_factors_range), 
				for ifeature=1:length(n_features_range), 
					for iweight=1:length(neg_weight_range), 

						params.model.alpha                    = alpha_range(ialpha);
						params.model.n_factors                = n_factors_range(ifactor);
						params.model.n_features               = n_features_range(ifeature);
						params.model.neg_relationships_weight = neg_weight_range(iweight);

						% Train the model.
						[S, O, U, V, alpha, Y, params] = tensor_factor(train_triplets(train_idx{ifold},:), train_related(train_idx{ifold}), params);

						% Test the model on the held-out split
						test_likelihood = get_likelihood_by_chunking(train_triplets(test_idx{ifold},:), train_related(test_idx{ifold}), S, O, U, V, alpha, Y, params);
					
						CV_scores(ifold, ialpha, ifactor, ifeature, iweight) = auc_pr(test_likelihood',train_related(test_idx{ifold})');
		
					end
				end
			end
		end
	end

	fprintf('\t\t*** %d-th random split (out of %d): %d-th fold (out of %d), testing on held-out set ***\n',s,n_splits,ifold,k_folds);

	% The matrix CV_scores has been completely filled, we now extract the best parameters
	dim            = size(CV_scores);
	mean_CV_scores = squeeze(mean(CV_scores,1));
	
	[best_score, best_idx] = max( mean_CV_scores(:) );

	[ialpha, ifactor, ifeature, iweight] = ind2sub(dim(2:end), best_idx);

	params.model.alpha                    = alpha_range(ialpha);
	params.model.n_factors                = n_factors_range(ifactor);
	params.model.n_features               = n_features_range(ifeature);
	params.model.neg_relationships_weight = neg_weight_range(iweight);

        % We store the best parameters
	best_params(s,:) = [params.model.alpha, params.model.n_factors, params.model.n_features, params.model.neg_relationships_weight];

	params.opt.max_iter = max_iter_test;
	% We retrain and test our model
	[S, O, U, V, alpha, Y, params] = tensor_factor(train_triplets, train_related, params);

	% For testing the extended nations dataset, we need to filter out the triplets which have been artificially added
	if strcmp(dataset,'dnations_extended')==1,
	
		filter        = (test_triplets(:,1) <= 14) & (test_triplets(:,2) <= 56) & (test_triplets(:,3) <= 14);	
		test_triplets = test_triplets(filter,:);
		test_related  = test_related(filter);

	end
	test_likelihood = get_likelihood_by_chunking(test_triplets, test_related, S, O, U, V, alpha, Y, params);

	params.opt.max_iter = max_iter_train;

	rd_split_scores(s,1) = auc_pr(test_likelihood',test_related');
	rd_split_scores(s,2) = mean(log(test_likelihood));
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[mean(rd_split_scores(:,1)), std(rd_split_scores(:,1))]% mean and standard deviation of the test AUC over the 10 random splits
[mean(rd_split_scores(:,2)), std(rd_split_scores(:,2))]% mean and standard deviation of the test mean log likelihood over the 10 random splits

file_name = [dataset, '_nsplits=', num2str(n_splits), '_kfolds=', num2str(k_folds), '_maxit=', num2str(max_iter_test), '_',  params.misc.exp_name, '.mat'];

save(file_name,'S','O','U','V','alpha','Y','params','rd_split_scores','best_params');
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
