% Provides a demo of the tensor_factor code.
% Trains a model on 90% of the dataset and validates it on the remaining 10%.
% The area under precision-recall curve is computed for both the training and the validation set.
%
% Copyright (C) 2012 Rodolphe Jenatton, Nicolas Le Roux, Antoine Bordes, Guillaume Obozinski
%         
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details. 
% 
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.


clear

dataset = 'uml';

switch dataset
case 'alyawarra'
	load('alyawarradata.mat');
case 'dnations'
	load('dnations.mat');
	Rs = R;
case 'uml'
	load('uml.mat');
end

triplets = zeros(numel(Rs), 3);
related = zeros(numel(Rs), 1);

% Transform the 3D-tensor into a matrix.
index = 0;
for i = 1:size(Rs, 1)
	for j = 1:size(Rs, 2)
		for k = 1:size(Rs, 3)
			% There might be NaNs.
			if ~isnan(Rs(i,j,k))
				index = index + 1;
				triplets(index, :) = [i k j];
				related(index, 1) = Rs(i,j,k);
			end
		end
	end
end

% Remove unused rows (which might be there because of the NaN).
triplets = triplets(1:index, :);
related = related(1:index, :);

% Compute the number of training and validation examples we will use.
n_train = floor(.9*rows(triplets));
n_valid = floor(.1*rows(triplets));

n_relationships = size(Rs, 3);
n_subjects = size(Rs, 1);
n_objects = n_subjects;

% Specify the parameters of the model
% Each parameter and its role are described in the file relational_params.m
params.opt.batchsize = 100;
params.opt.max_iter  = 10;
params.opt.rate      = .1;
params.opt.decr      = 0;
params.opt.proj_skip = 1;

params.misc.save = 1;

params.model.alpha                    = 0;
params.model.unique_rep               = 1;
params.model.constraint_for_SO        = 0;
params.model.n_factors                = 100;
params.model.n_features               = 10;
params.model.neg_relationships_weight = 1/2;
params.model.ngram                    = '111';

params.neg_sampling.sample_type = 'rel';
params.neg_sampling.n_neg       = 0;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% We need to make sure that the training set contains
% each subject, relationship and object at least once.
has_all = 0;
tries   = 0;

while has_all == 0
	R        = randperm(rows(triplets));
	triplets = triplets(R, :);
	related  = related(R, :);

	train_triplets = triplets(1:n_train, :);

	has_all = numel(unique(train_triplets(:, 2))) == n_relationships && numel(unique([train_triplets(:, 1);train_triplets(:, 3)])) == n_subjects;

	tries = tries + 1;

	if tries > 50, fprintf('Too many tries, aborted.\n'); return; end
end

train_related = related(1:n_train);

valid_triplets = triplets(n_train+(1:n_valid), :);
valid_related  = related(n_train+(1:n_valid));

% Train the model.
[S, O, U, V, alpha, Y, params] = tensor_factor(train_triplets, train_related, params);

% Compute the probabilities of each relation in the training and validation set.
train_probs = get_likelihood_by_chunking(train_triplets, train_related, S, O, U, V, alpha, Y, params);
train_auc   = auc_pr(train_probs', train_related');

valid_probs = get_likelihood_by_chunking(valid_triplets, valid_related, S, O, U, V, alpha, Y, params);
valid_auc   = auc_pr(valid_probs', valid_related');

fprintf('Area under precision-recall curve on the training set: %g\n', train_auc);
fprintf('Area under precision-recall curve on the validation set: %g\n', valid_auc);
