function projected_v = L1_proj( v, radius, IsPos)
% L1_PROJ 
%
% The function operates columnwise, i.e., each column of v is projected
% onto the simplex
%
% This function is currently in Matlab. If you care about efficiency,
% we encourage you to use mexSparseProject from the SPAMS toolbox.
% The code is provided below.
%
%% Use this code instead if you have the SPAMS toolbox installed.
%  param.pos        = IsPos;
%  param.mode       = 1;
%  param.thrs       = radius;
%  param.numThreads = 1;
%  
%  projected_v = mexSparseProject(v,param);
%
% Copyright (C) 2012 Rodolphe Jenatton, Nicolas Le Roux, Antoine Bordes, Guillaume Obozinski
%         
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details. 
% 
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.


projected_v = v;
if IsPos == 1
	% Project onto the probability simplex.
	for column = 1:cols(v)
		projected_v(:, column) = ProjectOntoSimplex(v(:, column), radius);
	end
else
	for column = 1:cols(v)
		projected_v(:, column) = ProjectOntoL1Ball(v(:, column), radius);
	end
end

end
%
% The following two functions are taken from http://www.cs.berkeley.edu/~jduchi/projects/DuchiShSiCh08.html
%
function w = ProjectOntoL1Ball(v, b)
	% PROJECTONTOL1BALL Projects point onto L1 ball of specified radius.
	%
	% w = ProjectOntoL1Ball(v, b) returns the vector w which is the solution
	%   to the following constrained minimization problem:
	%
	%    min   ||w - v||_2
	%    s.t.  ||w||_1 <= b.
	%
	%   That is, performs Euclidean projection of v to the 1-norm ball of radius
	%   b.
	%
	% Author: John Duchi (jduchi@cs.berkeley.edu)

	if (b < 0)
	  error('Radius of L1 ball is negative: %2.3f\n', b);
	end
	u = sort(abs(v),'descend');
	sv = cumsum(u);
	rho = find(u > (sv - b) ./ (1:length(u))', 1, 'last');
	theta = max(0, (sv(rho) - b) / rho);
	w = sign(v) .* max(abs(v) - theta, 0);
end


function w = ProjectOntoSimplex(v, b)
	% PROJECTONTOSIMPLEX Projects point onto simplex of specified radius.
	%
	% w = ProjectOntoSimplex(v, b) returns the vector w which is the solution
	%   to the following constrained minimization problem:
	%
	%    min   ||w - v||_2
	%    s.t.  sum(w) <= b, w >= 0.
	%
	%   That is, performs Euclidean projection of v to the positive simplex of
	%   radius b.
	%
	% Author: John Duchi (jduchi@cs.berkeley.edu)

	if (b < 0)
	  error('Radius of simplex is negative: %2.3f\n', b);
	end
	v = (v > 0) .* v;
	u = sort(v,'descend');
	sv = cumsum(u);
	rho = find(u > (sv - b) ./ (1:length(u))', 1, 'last');
	theta = max(0, (sv(rho) - b) / rho);
	w = max(v - theta, 0);
end
