//
// Copyright (C) 2012 Rodolphe Jenatton, Nicolas Le Roux, Antoine Bordes, Guillaume Obozinski
//         
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details. 
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
#include <mexutils.h>
#include "linalg.h"

/**************************************************************************
This function makes the computation in C++ of the following matlab code:

 grad_U = bsxfun(@plus, delta * A, Y(:,1)) * B' + Y(:,3) * sum(C,2)' + A * D';
 
 grad_V = bsxfun(@plus, delta * E, Y(:,2)) * F' + Y(:,4) * sum(G,2)' + E * H';
**************************************************************************/


void getGrad(double* OUT, bool delta, double* A, double* B, double* C, double* D, double* Y, const mwIndex* idx, mwSize n, mwSize nc, mwSize na)
{
    // OUT is assumed to have nc*na double numbers allocated
    mwIndex i, j;
    double* sum_row = new double[na];
    double* tmp     = new double[nc*n];
     
    //sum rows of C
    for (i=0;i<na;i++) sum_row[i] = 0.0;

    for (j=0;j<n;j++)
       for (i=0;i<na;i++) sum_row[i] += C[i+na*j];

    //
    for (j=0;j<na;j++)
        for (i=0;i<nc;i++)
            OUT[i+nc*j] = Y[i+nc*idx[1]] * sum_row[j];
    
    Matrix<double> matrix_OUT(OUT,nc,na); 
    Matrix<double> matrix_A(A,nc,n); 
    Matrix<double> matrix_B(B,na,n); 
    Matrix<double> matrix_D(D,na,n); 
    Matrix<double> matrix_tmp(tmp,nc,na);

    matrix_A.mult(matrix_D, matrix_OUT, false, true, 1.0, 1.0);	
    
    if (delta) 
        matrix_tmp.copy(matrix_A);
    else 
        for (i=0;i<nc*n;i++) tmp[i] = 0.0;
    
    for (j=0;j<n;j++)
        for (i=0;i<nc;i++)
            tmp[i+nc*j] += Y[i+nc*idx[0]];
    
    matrix_tmp.mult(matrix_B, matrix_OUT, false, true, 1.0, 1.0);
    
    delete[] sum_row;
    delete[] tmp;
}
//
// Argument 0: A,     whose dim is nc x n
//          1: B,     whose dim is na x n
//          2: C,     whose dim is na x n 
//          3: D,     whose dim is na x n 
//          4: Y      whose dim is nc x 4
//          5: idx    whose dim is 2 x 1
//          6:delta   scalar
//
//-------------------------------------------------------------------------
// mex  mexGetGrad_U_V.cpp  -largeArrayDims -lblas -llapack -I../mex/
//
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )
{
  
  double *A, *B, *C, *D, *Y, *idx, *d;
  
  mwSize nc = mxGetM(prhs[0]), n = mxGetN(prhs[0]), na = mxGetM(prhs[1]);

  A   = (double*)mxGetPr(prhs[0]);
  B   = (double*)mxGetPr(prhs[1]);
  C   = (double*)mxGetPr(prhs[2]);
  D   = (double*)mxGetPr(prhs[3]);
  Y   = (double*)mxGetPr(prhs[4]);
  idx = (double*)mxGetPr(prhs[5]);
  d   = (double*)mxGetPr(prhs[6]);
  
  bool delta       = *d;
  const mwIndex IDX[] = {idx[0], idx[1]}; 
  //-----------------------------------------------------------------------
  plhs[0]     = mxCreateDoubleMatrix(nc,na,mxREAL);
  double* OUT = (double*)mxGetPr(plhs[0]);

  getGrad(OUT, delta, A, B, C, D, Y, IDX, n, nc, na);
 
}

