function params = relational_params(params)

% relational_params makes sure that all the required fields in 'params' are set
% and sets them in the opposite case.
%
% The important values to set are:
% - params.model.n_features is the dimension of the representation of the subjects and objects.
% - params.model.n_factors is the number of latent factors used for the representation of relationships.
% - params.model.ngram is a string defining which terms are used. For instance, params.model.ngram = '011' means that bigrams and trigrams are used, but not unigrams.
% - params.model.neg_relationships_weight defines how much we scale the cost of the negative examples. This can be helpful in case of severe imbalance.
%
% Copyright (C) 2012 Rodolphe Jenatton, Nicolas Le Roux, Antoine Bordes, Guillaume Obozinski
%         
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details. 
% 
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

%%% PARAMETERS OF THE MODEL %%%

if ~isfield(params, 'model'), params.model = []; end

% Number of latent classes.
params.model = ask_params(params.model, 'n_features', 'pos_int', 'Number of latent classes (default = 10)? ', 10);

% Number of factors.
params.model = ask_params(params.model, 'n_factors', 'pos_int', ['Number of atoms (default = ', num2str(2*params.model.n_features), ')? '], 2*params.model.n_features);

% Do we force the matrices S and O to be equal?
params.model = ask_params(params.model, 'unique_rep', 'binary', ['Do we use the same matrix for the subject and the object (default = 1)? '], 1);

% Choose the constraints for the columns of S and O
% 0: L2 without positivity, 1: simplex (i.e., L1 with positivity), 2: L2 with positivity
params.model = ask_params(params.model, 'constraint_for_SO', 'pos_int', ['Choose the constraints for the columns of S and O; 0: L2 without pos, 1: simplex, 2: L2 with pos (default = 2)? '], 2);

% Which interaction terms do we consider? '101' means for instance uni and trigrams, but no bigrams. Similarly, '100' means only unigrams
% The first element is the unigram term, the second one is the bigram term and the third one is the trigram term.
params.model = ask_params(params.model, 'ngram', 'string', ['Which interaction terms do we consider (default = "111", i.e., uni-, bi- and tri-gram)? '], '111');

%% How much do we scale the cost for each negative example?
params.model = ask_params(params.model, 'neg_relationships_weight', 'pos', 'Scaling of the cost for the negative examples? (by default = 1)', 1);

% Constraint on alpha.
params.model = ask_params(params.model, 'alpha', 'real', 'Constraint on alpha. Positive value is L1 constraint, negative value is L2 constraint (default = -1)? ', -1);


%%% PARAMETERS OF THE OPTIMISATION PROCEDURE %%%

if ~isfield(params, 'opt'), params.opt = []; end

%% Number of iterations through the entire training set.
params.opt = ask_params(params.opt, 'max_iter', 'pos_int', 'How many iterations through the training set (default = 20)? ', 20);

% Size of the minibatches.
params.opt = ask_params(params.opt, 'batchsize', 'pos_int', 'Size of the mini-batches (default = 100)? ', 100);

% Learning rate.
params.opt = ask_params(params.opt, 'rate', 'pos', 'Learning rate (default = .5)? ', .2);

% Decrease constant.
params.opt = ask_params(params.opt, 'decr', 'pos', 'Decrease constant (default = 0)? ', 0);

% How many steps to skip between two projections on the L2-ball?
params.opt = ask_params(params.opt, 'proj_skip', 'pos_int', 'Every how many steps do we project the parameters onto the L2-ball (default = 1 = always)? ', 1);


% Params related to the sampling of the negative triplets
if ~isfield(params, 'neg_sampling'), params.neg_sampling = []; end

params.neg_sampling = ask_params(params.neg_sampling, 'n_neg', 'pos', 'Number of negative triplets to generate, as a fraction of the dataset size (default = 0)? ', 0);

params.neg_sampling = ask_params(params.neg_sampling, 'sample_type', 'string', 'Define the sampling scheme (default = ''rel'')? ', 'rel');

% Various options.
if ~isfield(params, 'misc'), params.misc = []; end
% Save the model every params.misc.save epochs (0 = disable).
params.misc = ask_params(params.misc, 'save', 'pos_int', 'Every how many epochs do we save the model (default = 0 = disable)? ', 0);

if ~isfield(params.misc, 'exp_name'), 
    params.misc.exp_name = [...
	'features=', num2str(params.model.n_features), '_', ...
	'factors=', num2str(params.model.n_factors),'_', ...
    'neg_relationships_weight=', num2str(params.model.neg_relationships_weight),'_',...
    'alpha=', num2str(params.model.alpha),'_',...
    'ngram=', params.model.ngram]; 
end

% Path to save the file.
if ~isfield(params.misc, 'path'), params.misc.path = '.'; end
