function [S, O, U, V, alpha, Y, params] = tensor_factor(triplets, related, params, restart_values)
%
%
%%% OUTPUTS %%%
%
% - S: the matrix containing the representation of the subjects,              size: n_features x n_subjects
% - O: the matrix containing the representation of the objects,               size: n_features x n_objects
% - U: the first matrix containing the representation of the relationships,   size: n_features x n_factors
% - V: the second matrix containing the representation of the relationships,  size: n_features x n_factors
% - alpha: the matrix containing the coefficients for the relationship atoms, size: n_factors x n_relationships
%     - Representation of relationship k is \sum_i alpha(i, k) U(:, i) V(:, i)^T.
% - Y: the matrix containing the weight vectors for the uni and bigram terms, size: n_features x 4
%     - Y = [Y_s, Y_o, Y_{rel_s}, Y_{rel_o}] includes the terms "z" from the NIPS paper.
%
%%% INPUTS %%%
%
% - triplets: contains the indices [i, j, k] of a relationship for subject i, relation j, and object k,           size: n_triplets x 3
% - related: is a {0,1} vector determining whether the corresponding relationships in triplets exist,             size: n_triplets x 1
%     - related(a) == 1 means that the a-th relationship in triplets---that is, triplets(a,:)---exists in the dataset
%     - by default, if lenght(related) < rows(triplets), then related is assumed to be equal to ones(n_triplets,1) 
% - params: is a set of parameters for the algorithm. Relevant fields and default values are stored in relation_params.m .
% - restart_values: is an optional parameter which enables a warm restart of the algorithm.
%
% Copyright (C) 2012 Rodolphe Jenatton, Nicolas Le Roux, Antoine Bordes, Guillaume Obozinski
%         
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details. 
% 
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.


% We check that some data points are provided
if size(triplets,1) == 0,
	error('There is no data points provided!');
end

%We make sure that related is of the proper size:
if length(related) < size(triplets,1),
	related = ones(size(triplets,1),1);
end
%
% We only want to keep the relationships which appear at least once in the dataset.
% Same for the subjects and the objects.
%
% subjects are in the first column of "triplets"
% relationships are in the second column of "triplets"
% objects are in the third column of "triplets"

% unique_subjects contains the list of IDs for the subjects.
unique_subjects = unique(triplets(:, 1));

% unique_rel contains the list of IDs for the relationships.
unique_rel = unique(triplets(:, 2));

% unique_objects contains the list of IDs for the objects.
unique_objects = unique(triplets(:, 3));

% The indices in "triplets" may be arbitrary.
% We thus need to remap them to indices from 1 to the number of different elements.

%% RELATIONSHIPS %%

% Compute and store the total number of different relationships.
n_relationships = numel(unique_rel);
params.n_relationships = n_relationships;

% list_relationships(old_ID) contains the new index associated with the old ID old_ID.
list_relationships(unique_rel) = 1:n_relationships;

% We may now replace all the old IDs with the new ones.
triplets(:, 2) = list_relationships(triplets(:, 2));

% We must also store the map between original and new indices.
params.misc.relationship_indices = list_relationships;

% If we use a unique representation for subjects and objects, they need to have a common index.
if params.model.unique_rep == 1,

	unique_subjects_objects                = unique([unique_subjects;unique_objects]);
	n_subjects                             = numel(unique_subjects_objects);
	n_objects                              = n_subjects;
	list_subjects(unique_subjects_objects) = 1:n_subjects;
	list_objects                           = list_subjects;
else

	%% SUBJECTS %%
	n_subjects = numel(unique_subjects);

	% list_subjects(old_ID) contains the new index associated with the old ID old_ID.
	list_subjects(unique_subjects) = 1:n_subjects;

	%% OBJECTS %%
	n_objects = numel(unique_objects);
	
	% list_objects(old_ID) contains the new index associated with the old ID old_ID.
	list_objects(unique_objects) = 1:n_objects;

end

% We may now replace all the old IDs with the new ones for the subjects and the objects.
triplets(:, 1) = list_subjects(triplets(:, 1));
triplets(:, 3) = list_objects(triplets(:, 3));

% We must also return the map between original and new indices.
params.misc.subject_indices = list_subjects;
params.misc.object_indices = list_objects;

% If we do not do a warm restart.
if nargin < 4

	params = relational_params(params);

	% Scale appropriately the regularization when using the l1 norm
	if params.model.alpha > 0
	    params.model.alpha_scaled = params.model.alpha * n_relationships * params.model.n_factors;
	end
    
	% This is the size of the representation for each subject and each object.
	n_features = params.model.n_features;

	% This is the number of factors for the relationship matrices.
	n_factors = params.model.n_factors;

	% Maximum number of passes through the dataset.
	max_iter = params.opt.max_iter;

	% Initialize the matrices S and O.
	% They are of size n_features by n_subjects and n_features by n_objects.
    S = rand(n_features, n_subjects);

    % Do we constrain S and O to be equal?
    if params.model.unique_rep == 1
    	O = S;
    else
    	O = rand(n_features, n_objects);
   	end
    
    % We now appropriately normalize S and O.
    switch params.model.constraint_for_SO
        case 0, % Project onto L2 ball
            S = L2_proj(S,0);
            O = L2_proj(O,0);
        case 1, % Project onto simplex
            S = L1_proj(S,1,1);
            O = L1_proj(O,1,1);
        case 2, % Project onto L2 ball with pos constraints
            S = L2_proj(S,1);
            O = L2_proj(O,1);
        otherwise
            S = L2_proj(S,0);
            O = L2_proj(O,0);
    end

	% Initialize the matrices U and V which are the rank-one components of the relationship matrices.
	% They are of size n_features by n_factors.
	% alpha is of size n_relationships by n_factors.

	% NLR: changed rand in randn. Check that is does not have any influence.

    U     = randn(n_features, n_factors);
    U     = L2_proj(U,0);

    V     = randn(n_features, n_factors);
    V     = L2_proj(V,0);

    alpha = randn(n_factors, n_relationships);
    
    % Y corresponds to the weight vectors for the uni and bigram terms Y = [Y_s, Y_o, Y_{rel_s}, Y_{rel_o}]
    % Y includes the terms "z" from the NIPS paper.
    Y = zeros(n_features,4); 

    % Initialize the counter which determines when to do the projection of the matrices onto their constraint set.
	params.opt.skip_step = 1;

	% Getting the time.
	temp_clock        = clock;
	year              = temp_clock(1);
	month             = temp_clock(2);
	day               = temp_clock(3);
	hour              = temp_clock(4);
	minute            = temp_clock(5);
	start_time        = sprintf('%i_%i_%i_%i_%i',year,month,day,hour,minute);
	params.start_time = start_time;

	start_iter  = 1;
	params.cost = zeros(1, max_iter);

else

	% We do a warm restart and need to retrieve the appropriate values.

	start_time = params.start_time;
	S          = restart_values.S;
	O          = restart_values.O;
	U          = restart_values.U;
	V          = restart_values.V;
	alpha      = restart_values.alpha;
	Y          = restart_values.Y;
	max_iter   = params.opt.max_iter;
	start_iter = params.current_iter + 1;

end

file_name = [start_time, '_', params.misc.exp_name];
dir_name = [params.misc.path, '/', file_name];

% Full size of the dataset: number of columns of triplets added to the number (possibly equal to 0) of negative triples generated afterwards
n_neg_relationships = floor(params.neg_sampling.n_neg * rows(triplets));
n_triplets = rows(triplets) + n_neg_relationships;

% Size of the minibatches.
batchsize = params.opt.batchsize;
n_batches = floor(n_triplets/batchsize);

% Storing the cost for all batches.
all_costs = zeros(n_batches, 1);

% We wish to make the choice of the learning rate as independent of the batchsize
% and of the scaling of negative examples as possible.
% Since we compute the gradient of the SUM of the costs over a minibatch, we will divide
% the learning rate by the average sum of cost weights across a minibatch.
params.opt.epsilon_correction = ...
	( mean(related) + params.model.neg_relationships_weight*mean(1 - related) ) * batchsize;

% Provides an estimate of the remaining time.
tic

% Starting the optimization.
for iter = start_iter:max_iter

	index = 0;
	params.current_iter = iter;

	triplets_to_add = generate_neg_relationships(triplets, n_neg_relationships, params.neg_sampling.sample_type);% possibly empty 	
	related_to_add  = zeros(n_neg_relationships,1);

	triplets_full = [triplets; triplets_to_add];
        related_full  = [related; related_to_add];

	% We randomize the order in which we treat the examples for each iteration.
	R = randperm(n_triplets);

	for batches = 1:n_batches

		% Be careful for the last batch, it may be bigger than the others.
		if batches == n_batches
			triplets_batch = triplets_full(R( (index+1) : end ), :);
			related_batch  = related_full(R( (index+1) : end ));
		else
			triplets_batch = triplets_full(R( index + (1:batchsize) ), :);
			related_batch  = related_full(R( index + (1:batchsize) ));
		end
		index = index + batchsize;

		% Compute the cost and the gradient with respect to the parameters.
		[cost_batch, grad] = tensor_factor_evaluation(triplets_batch, related_batch, S, O, U, V, alpha, Y, params);

		% Update the parameters.
		[S, O, U, V, alpha, Y, params] = tensor_factor_update(S, O, U, V, alpha, Y, grad, params);

		all_costs(batches) = cost_batch;

	end

	params.cost(iter) = sum(all_costs)/n_triplets;
	if ~mod(iter, params.misc.save)
        % If the directory does not exist, create it.
		if ~exist(dir_name, 'dir'), mkdir(dir_name); end
        save([dir_name, '/', file_name, '_iter=', num2str(iter), '_tensor_factor.mat'], 'S', 'O', 'U', 'V', 'alpha', 'Y', 'params');
    end
	time_str = sec_to_time(toc*(max_iter - iter)/(iter - start_iter + 1));
	fprintf('End of iteration %i/%i		Cost: %g		Time remaining: %s\n', iter, max_iter, params.cost(iter), time_str);

end
