function [cost, grad_S, grad_O, grad_U, grad_V, grad_alpha, grad_Y] =...
		tensor_factor_costgrad(data, related, S, O, U, V, alpha, Y, params)

% tensor_factor_costgrad computes the log-probability and the gradient
% with respect to all the parameters for all the triples (subject, relationship, object)
% contained in (data, related).
%
% Copyright (C) 2012 Rodolphe Jenatton, Nicolas Le Roux, Antoine Bordes, Guillaume Obozinski
%         
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details. 
% 
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

if params.model.ngram(3) == '1' %There is the tri-gram term
  is_trigram = 1;
else
  is_trigram = 0;
end

% rep_S only contains the representations of the subjects in the minibatch.
% rep_S is of size n_features by n_data.

% rep_O only contains the representations of the objects in the minibatch.
% rep_O is of size n_features by n_data.

rep_S     = S(:, data(:, 1));
rep_O     = O(:, data(:, 3));
rep_alpha = alpha(:, data(:,2));

SV_rel = U' * rep_S;
VO_rel = V' * rep_O;

VO_alpha = VO_rel .* rep_alpha;
SV_alpha = SV_rel .* rep_alpha;

% Y = [ Y_s, Y_o, Y_vs, Y_vo ]
U_Y_s  = U' * Y(:,1);
V_Y_o  = V' * Y(:,2);
U_Y_vs = U' * Y(:,3);
V_Y_vo = V' * Y(:,4);

energy = -V_Y_o' * SV_alpha - ...
         U_Y_s' * VO_alpha - ...
         ( V_Y_vo .* U_Y_vs )' * rep_alpha;

if is_trigram == 1
    energy = energy - sum(  SV_rel .* VO_alpha , 1);
end

energy = energy';

% We scale the cost of negative examples (related == 0) by params.model.neg_relationships_weight.
is_neg_relationship = (related == 0);
cost = -related.*energy + softplus(energy);
cost(is_neg_relationship) = cost(is_neg_relationship)*params.model.neg_relationships_weight;

% Only compute the gradient if there is more than one output.
if nargout > 1

    % If the examples considered are positive, the gradient for each datapoint must be multiplied by sigm(energy).
    % Otherwise, it must be multiplied by sigm(energy) - 1.
    weight = related' + sigm(-energy)' - 1;

    % The gradient of the negative examples is also scaled by params.model.neg_relationships_weight.
    weight(is_neg_relationship) = weight(is_neg_relationship)*params.model.neg_relationships_weight;

    VO_alpha_weighted = bsxfun(@times, VO_alpha, weight);
    SV_alpha_weighted = bsxfun(@times, SV_alpha, weight); 

    grad_S = U * VO_alpha_weighted;
    grad_O = V * SV_alpha_weighted;

    % Computation for Y
    grad_Y = zeros(size(Y));

    alpha_weighted = bsxfun(@times, rep_alpha, weight);

    alpha_weighted_V_Y_vo = bsxfun(@times, V_Y_vo, alpha_weighted);
    alpha_weighted_U_Y_vs = bsxfun(@times, U_Y_vs, alpha_weighted);

    grad_Y(:,1) = sum(grad_S,2);
    grad_Y(:,2) = sum(grad_O,2);
    grad_Y(:,3) = sum( U * alpha_weighted_V_Y_vo, 2);
    grad_Y(:,4) = sum( V * alpha_weighted_U_Y_vs, 2);    
    
    alpha_weighted_V_Y_o = bsxfun(@times, V_Y_o, alpha_weighted);
    alpha_weighted_U_Y_s = bsxfun(@times, U_Y_s, alpha_weighted);

    if is_trigram == 1
        grad_S = grad_S + U * alpha_weighted_V_Y_o;
        grad_O = grad_O + V * alpha_weighted_U_Y_s;   
    else
        grad_S = U * alpha_weighted_V_Y_o;
        grad_O = V * alpha_weighted_U_Y_s;          
    end
    
    grad_U     = mexGetGrad_U_V(rep_S, VO_alpha_weighted, alpha_weighted_V_Y_vo, alpha_weighted_V_Y_o, Y, [0 2], is_trigram);
    grad_V     = mexGetGrad_U_V(rep_O, SV_alpha_weighted, alpha_weighted_U_Y_vs, alpha_weighted_U_Y_s, Y, [1 3], is_trigram); 
    grad_alpha = mexGetGrad_alpha(SV_rel, VO_rel, V_Y_o, U_Y_s, V_Y_vo, U_Y_vs, weight, is_trigram);         
end
