function [cost, grad, all_costs] = tensor_factor_evaluation(data, related, S, O, U, V, alpha, Y, params, is_test)
%
% The function tensor_factor_evaluation gathers the computations of the cost and gradient for the mini-batch given by {data, related}
%
% Copyright (C) 2012 Rodolphe Jenatton, Nicolas Le Roux, Antoine Bordes, Guillaume Obozinski
%         
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details. 
% 
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

n_verbs    = cols(alpha);
n_data     = rows(data);

% is_test is 1 when we are testing the model and do not need to compute the gradient.
if nargin < 10, is_test = 0; end

% These matrices only store the gradients for the examples in the minibatch.
% These gradients will then be used to fill the proper gradient matrices using a C++ routine.
if is_test == 0
	grad_O_temp     = zeros(rows(O), n_data);
	grad_S_temp     = zeros(rows(S), n_data);
	grad_alpha_temp = zeros(rows(alpha), n_data);
end

% Compute the cost.
if is_test == 0
	[all_costs, grad_S_temp, grad_O_temp, grad_U, grad_V, grad_alpha_temp, grad_Y] = ...
		tensor_factor_costgrad(data, related, S, O, U, V, alpha, Y, params);
else
	all_costs = tensor_factor_costgrad(data, related, S, O, U, V, alpha, Y, params);
end

% Cost is the sum of all costs for that minibatch.
cost = sum(all_costs);

if is_test == 0
	% Build the sparse gradient for O, S and alpha

	grad_alpha = mexCreateSpGrad( data(:, 2), grad_alpha_temp, 1, cols(alpha));
	grad_S     = mexCreateSpGrad( data(:, 1), grad_S_temp, 1, cols(S));
	grad_O     = mexCreateSpGrad( data(:, 3), grad_O_temp, 1, cols(O));

	% Regularization terms.
	if params.model.unique_rep == 1
		grad.S = grad_S + grad_O;
	else
		grad.S = grad_S;
		grad.O = grad_O;
	end

	grad.U     = grad_U;
	grad.V     = grad_V;
	grad.alpha = grad_alpha;
	grad.Y     = grad_Y;
end

% If it is the test, we have to create a dummy variable grad.
if is_test == 1, grad = 0; end

end
